@@ -14,6 +14,7 @@ playDevList <- function() {
 }
 
 playDevSet <- function(playState) {
+	stopifnot(inherits(playState, "playState"))
 	StateEnv$.current <- playState
 	playState$old.dev <- dev.cur()
 	dev.set(playState$dev)
@@ -25,8 +26,15 @@ playDevOff <- function(playState = playDevCur()) {
 	cleanupStateEnv()
 }
 
+print.playState <- function(x, ...) {
+	stopifnot(inherits(x, "playState"))
+	cat(paste("<playState: ", 
+		toString(x$win["title"]), ">\n", sep=""))
+}
+
 cleanupStateEnv <- function() {
 	for (ID in ls(StateEnv)) {
+		if (!inherits(StateEnv[[ID]], "playState")) next
 		if (!inherits(StateEnv[[ID]]$win, "GtkWindow")) {
 			# window is defunct
 			rm(list=ID, envir=StateEnv)
@@ -280,6 +288,11 @@ playSelectData <- function(playState, prompt="Click or drag to select data point
 	if (is.null(foo$coords)) return(NULL)
 	coords <- foo$coords
 	data <- xyCoords(playState, space=foo$space)
+	if (length(data$x) == 0) {
+		gmessage.error(paste("Sorry, can not guess the data point coordinates.",
+			"Please contact the maintainer with suggestions."))
+		return(NULL)
+	}
 	which <- NULL
 	pos <- NULL
 	if (foo$is.click) {
@@ -362,7 +375,9 @@ playLineInput <- function(playState, prompt="Click and drag to define a line") {
 	playClickOrDrag(playState, x0=xy0$x, y=xy0$y, shape="line")
 }
 
-playRectInput <- function(playState, prompt="Click and drag to define a rectangular region") {
+playRectInput <- function(playState, prompt="Click and drag to define a rectangular region",
+	scales=c("x", "y")) {
+	scales <- match.arg(scales, several.ok=TRUE)
 	playDevSet(playState)
 	playPrompt(playState, prompt)
 	on.exit(playPrompt(playState, NULL))
@@ -373,15 +388,15 @@ playRectInput <- function(playState, prompt="Click and drag to define a rectangu
 	xy0 <- grid.locator()
 	if (is.null(xy0)) return(NULL)
 	xy0 <- lapply(xy0, as.numeric)
-	playClickOrDrag(playState, x0=xy0$x, y=xy0$y, shape="rect")
+	playClickOrDrag(playState, x0=xy0$x, y=xy0$y, shape="rect", scales=scales)
 }
 
 # assumes that the mouse button has already been pressed
 # converts into user coordinates
-playClickOrDrag <- function(playState, x0, y0, shape=c("rect", "line")) {
-	shape <- match.arg(shape)
-	da <- playState$widgets$drawingArea
-	foo <- handleClickOrDrag(da, x0=x0, y0=y0, shape=shape)
+playClickOrDrag <- function(playState, x0, y0, shape=c("rect", "line"),
+	scales=c("x", "y")) {
+	foo <- handleClickOrDrag(playState$widgets$drawingArea, 
+		x0=x0, y0=y0, shape=shape, scales=scales)
 	if (is.null(foo)) return(NULL)
 	dc <- foo$dc
 	coords <- NULL
@@ -402,8 +417,11 @@ playClickOrDrag <- function(playState, x0, y0, shape=c("rect", "line")) {
 }
 
 # assumes that the mouse button has already been pressed
-handleClickOrDrag <- function(da, x0, y0, shape=c("rect", "line")) {
+handleClickOrDrag <- function(da, x0, y0, shape=c("rect", "line"),
+	scales=c("x", "y")) {
 	shape <- match.arg(shape)
+	scales <- match.arg(scales, several.ok=TRUE)
+	# px0 is the original click location
 	px0 <- list(x=x0, y=y0)
 	da.w <- da$getAllocation()$width
 	da.h <- da$getAllocation()$height
@@ -431,6 +449,9 @@ handleClickOrDrag <- function(da, x0, y0, shape=c("rect", "line")) {
 			width=area$width, height=area$height)
 		xx <- range(c(env$px0$x, env$px00$x))
 		yy <- range(c(env$px0$y, env$px00$y))
+		# restrict rectangle to x or y scales, according to `scales`
+		if (!("x" %in% scales)) xx <- c(-1, da.w)
+		if (!("y" %in% scales)) yy <- c(-1, da.h)
 		wd <- xx[2] - xx[1]
 		ht <- yy[2] - yy[1]
 		switch(env$shape,
@@ -445,6 +466,7 @@ handleClickOrDrag <- function(da, x0, y0, shape=c("rect", "line")) {
 	tmpSigE <- gSignalConnect(da, "expose-event", expose_handler, data=environment())
 	tmpSigR <- gSignalConnect(da, "button-release-event", release_handler, data=environment())
 	repeat {
+		# px1 is the final drag location, set by event handler
 		if (exists("px1", inherits=FALSE)) break
 		px00.prev <- px00
 		px00 <- da$window$getPointer()
@@ -456,6 +478,9 @@ handleClickOrDrag <- function(da, x0, y0, shape=c("rect", "line")) {
 		}
 		xx <- range(c(px0$x, px00$x, px00.prev$x))
 		yy <- range(c(px0$y, px00$y, px00.prev$y))
+		# restrict rectangle to x or y scales, according to `scales`
+		if (!("x" %in% scales)) xx <- c(-1, da.w)
+		if (!("y" %in% scales)) yy <- c(-1, da.h)
 		wd <- xx[2] - xx[1] + 2
 		ht <- yy[2] - yy[1] + 2
 		da$window$invalidateRect(list(x=xx[1], y=yy[1], width=wd, height=ht),
@@ -506,6 +531,10 @@ xyData <- function(playState, space="plot") {
 		}
 		packet <- as.numeric(sub("packet ", "", space))
 		foo <- trellis.panelArgs(playState$trellis, packet.number=packet)
+		if (is.null(foo$y) && !is.null(foo$distribution)) {
+			# probably `qqmath`
+			return(xy.coords.qqmath(panel.args=foo))
+		}
 		if (length(foo$x) != length(foo$y)) {
 			if ((nx <- length(foo$x)) < (ny <- length(foo$y))) 
 				foo$x <- rep(foo$x, length.out = ny)
@@ -578,6 +607,52 @@ xy.coords_with_class <- function(x, y=NULL, recycle=TRUE) {
 	list(x=x, y=y)
 }
 
+# copied from panel.identify.qqmath
+xy.coords.qqmath <-
+    function(x = panel.args$x,
+             distribution = panel.args$distribution,
+             groups = panel.args$groups, 
+             subscripts = panel.args$subscripts,
+             labels = subscripts, 
+             panel.args = trellis.panelArgs(),
+             ...)
+{
+    x <- as.numeric(x)
+    if (is.null(subscripts)) subscripts <- seq_along(x)
+    if (!is.null(panel.args$f.value)) warning("'f.value' not supported; ignoring")
+    distribution <-
+        if (is.function(distribution)) distribution 
+        else if (is.character(distribution)) get(distribution)
+        else eval(distribution)
+    ## compute quantiles corresponding to given vector, possibly
+    ## containing NA's.  The return value must correspond to the
+    ## original order
+    getq <- function(x)
+    {
+        ans <- x
+        id <- !is.na(x)
+        ord <- order(x[id])
+        if (any(id)) ans[id] <- distribution(ppoints(sum(id)))[order(ord)]
+        ans
+    }
+    if (is.null(groups))
+    {
+        return(list(x = getq(x), y = x, subscripts = subscripts))
+    }
+    else
+    {
+        allq <- rep(NA_real_, length(x))
+        subg <- groups[subscripts]
+        vals <- if (is.factor(groups)) levels(groups) else sort(unique(groups))
+        for (i in seq_along(vals))
+        {
+            ok <- !is.na(subg) & (subg == vals[i])
+            allq[ok] <- getq(x[ok])
+        }
+        return(list(x = allq, y = x, subscripts = subscripts))
+    }
+}
+
 unlogX <- function(x, the.call, is.lattice=T) {
 	unlogXY(x, the.call, is.lattice, x.or.y="x")
 }
@@ -648,7 +723,7 @@ deviceToUserCoordsFunction <- function(is.grid = TRUE) {
 		diff(convertY(unit(0:1, "npc"), "native", valueOnly=TRUE))))
 	# pixels per inch
 	dpi <- convertX(unit(1, "inches"), "native", valueOnly=TRUE)
-	if (!is.null(vp)) downViewport(vp)
+	if (length(vp) > 0) downViewport(vp)
 	# device size in inches
 	din <- dpx / dpi
 	if (is.grid) {
